
import random
import matplotlib.colors as mcolors
import matplotlib.pyplot as plt
import networkx as nx
import numpy as np
import pandas as pd
import seaborn as sns
from scipy.interpolate import make_interp_spline, BSpline
from sklearn.linear_model import LinearRegression
from sklearn.preprocessing import PolynomialFeatures

plt.switch_backend('TkAgg')
plt.rcParams['font.sans-serif']=['SimHei']
plt.rcParams['axes.unicode_minus']=False

# 可视化辅助函数
def draw(G, pos, measures, measure_name):
    plt.figure(figsize=(20, 20))
    nodes = nx.draw_networkx_nodes(G, pos, node_size=250, cmap=plt.cm.plasma,
                                   node_color=list(measures.values()),
                                   nodelist=measures.keys())
    nodes.set_norm(mcolors.SymLogNorm(linthresh=0.01, linscale=1, base=10))
    # labels=nx.draw_networkx_labels(G,pos)
    edges = nx.draw_networkx_edges(G, pos)

    plt.title(measure_name, fontsize=30)
    # plt.colorbar(nodes)
    plt.axis('off')
    plt.show()


def dict_sort_by_value(dict_input):
    '''
    输入字典，输出按值排序的字典
    '''
    return sorted(dict_input.items(), key=lambda x:x[1], reverse=True)


df = pd.read_csv('gongjiao.csv',encoding='gbk')



G = nx.Graph()


for idx, row in df.iterrows():
    G.add_edges_from([(row['Previous stop'], row['Next stop'])], line=row['Subway line'], time=row['Time(minutes)'])

pos = nx.spring_layout(G, seed=123)


degree = dict(G.degree())
average_neighbor_degree = nx.average_neighbor_degree(G)


deg_values = list(degree.values())
avg_neighbor_deg_values = [average_neighbor_degree[node] for node in degree]


plt.figure(figsize=(8, 6))
plt.scatter(deg_values, avg_neighbor_deg_values, color='b', alpha=0.5, edgecolors='k', s=50)
plt.title('Degree vs Average Neighbor Degree', fontsize=16)
plt.xlabel('Degree', fontsize=14)
plt.ylabel('Average Neighbor Degree', fontsize=14)
plt.grid(True)
plt.show()

def compute_resilience(graph, nodes_to_remove):
    remaining_graph = graph.copy()
    remaining_graph.remove_nodes_from(nodes_to_remove)
    if len(remaining_graph.nodes()) > 0:
        largest_cc = max(nx.connected_components(remaining_graph), key=len)
        return len(largest_cc)
    else:
        return 0


nodes = list(G.nodes())
num_nodes = len(nodes)
num_removed = []
resilience_levels = []

for i in range(num_nodes):
    random_nodes = random.sample(nodes, i+1)
    num_removed.append(i+1)
    resilience_levels.append(compute_resilience(G, random_nodes))


def smooth_data(data, window_size=3):
    return np.convolve(data, np.ones(window_size)/window_size, mode='valid')

smoothed_resilience = smooth_data(resilience_levels)


plt.plot(num_removed[:len(smoothed_resilience)], smoothed_resilience, marker='o')
plt.xlabel('Number of Nodes Removed')
plt.ylabel('Resilience Level (Size of Largest Connected Component)')
plt.title('Network Resilience')
plt.grid(True)
plt.show()


def compute_resilience(graph, nodes_to_remove):
    remaining_graph = graph.copy()
    remaining_graph.remove_nodes_from(nodes_to_remove)
    if len(remaining_graph.nodes()) > 0:
        largest_cc = max(nx.connected_components(remaining_graph), key=len)
        return len(largest_cc)
    else:
        return 0


nodes = list(G.nodes())
num_nodes = len(nodes)
num_removed = []
resilience_levels = []

for i in range(num_nodes):
    nodes_to_remove = nodes[:i+1]
    num_removed.append(i+1)
    resilience_levels.append(compute_resilience(G, nodes_to_remove))


x = np.array(num_removed)
y = np.array(resilience_levels)
x_new = np.linspace(x.min(), x.max(), 300)
spl = make_interp_spline(x, y, k=3)
y_smooth = spl(x_new)

def loess_smooth(x, y, span=0.3):
    x = np.array(x)
    y = np.array(y)
    span = int(len(x) * span)
    smoothed = np.zeros_like(y)
    for i in range(len(x)):
        weights = np.exp(-((x - x[i]) / (span / 2)) ** 2)
        weights /= weights.sum()
        smoothed[i] = np.dot(weights, y)
    return smoothed

y_loess = loess_smooth(x, y)

percentages = [(i+1) / num_nodes * 100 for i in range(num_nodes)]
x = np.array(percentages)



plt.figure(figsize=(12, 6))
plt.plot(x, y_loess, label='Composite Network',color='deepskyblue', linewidth=2, marker='o', markersize=4)
plt.xlabel('Percentage of Nodes Removed')
plt.ylabel('P')
plt.legend()
plt.savefig('tanxing3.png', dpi=300, bbox_inches='tight')
plt.grid(True, linestyle='--', alpha=0.7)
plt.show()


def compute_average_degree(graph):
    if len(graph.nodes()) > 0:
        degrees = [degree for node, degree in graph.degree()]
        return np.mean(degrees)
    else:
        return 0


nodes = list(G.nodes())
num_nodes = len(nodes)
percentages = np.linspace(0, 1, num_nodes)
avg_degrees = []


for p in percentages:
    num_removed = int(p * num_nodes)
    nodes_to_remove = nodes[:num_removed]
    remaining_graph = G.copy()
    remaining_graph.remove_nodes_from(nodes_to_remove)
    avg_degrees.append(compute_average_degree(remaining_graph))


plt.figure(figsize=(12, 6))
plt.plot(percentages * 100, avg_degrees, label='Bus Network', color='darkviolet', linewidth=2, marker='o', markersize=4)



plt.xlabel('Percentage of Nodes Removed')
plt.ylabel('Pv')
plt.legend()
plt.savefig('cuiruoxing2.png', dpi=300, bbox_inches='tight')
plt.grid(True, linestyle='--', alpha=0.7)
plt.show()

'''color_map = []
for node in G:
    if node == '金台路':
        color_map.append('red')
    elif node == '西直门':
            color_map.append('blue')
    else:
        color_map.append('yellow')

plt.figure(figsize=(15, 15))
nx.draw(G, pos=pos, node_color=color_map, node_size=50, with_labels=False, width=1)
plt.show()'''





# largest_connected_component = max(nx.connected_components(G), key=len)
# largest_connected_subgraph = G.subgraph(largest_connected_component)


# plt.figure(figsize=(15, 15))
# nx.draw(largest_connected_subgraph, node_color='yellow', node_size=50, with_labels=False, width=1)

# plt.show()



#print('天安门东到天安门西是否有路:', nx.has_path(G, source='天安门东', target='天安门西'))


#print('古城到天安门西的最短路径:', nx.shortest_path(G, source='古城', target='天安门西', weight='time'))


#print('古城到天安门西的最短时间:', nx.shortest_path_length(G, source='古城', target='天安门西', weight='time'))


# print('全图平均最短路径:', nx.average_shortest_path_length(G, weight='time'))


# print('军事博物馆去其他站的最短路径：', nx.single_source_shortest_path(G, source='军事博物馆'))


#print('军事博物馆去其他站的最短路径长度：', nx.single_source_shortest_path_length(G, source='军事博物馆'))


"""shortest_path_list = nx.shortest_path(G, source='丰台站', target='奥林匹克公园', weight='time')
for i in range(len(shortest_path_list)-1):
    previous_station = shortest_path_list[i]
    next_station = shortest_path_list[i+1]
    line_id = G.edges[(previous_station),(next_station)]['line'] #地铁线编号
    time = G.edges[(previous_station),(next_station)]['time']    #时间
    print('{}--->{} {}号线 {}分钟'.format(previous_station, next_station, line_id, time))"""


# print('共计{}分钟'.format(nx.shortest_path_length(G, source='五道口', target='天安门西', weight='time')))


path_lengths = dict(nx.all_pairs_shortest_path_length(G))


avg_shortest_path_lengths = []
for node in G.nodes():
    path_lengths = nx.shortest_path_length(G, source=node)
    avg_length = np.mean(list(path_lengths.values()))-9
    avg_shortest_path_lengths.append(avg_length)


avg_shortest_path_lengths.sort()
cumulative_distribution = np.arange(1, len(avg_shortest_path_lengths) + 1) / len(avg_shortest_path_lengths)


plt.figure(figsize=(8, 6))
plt.plot(avg_shortest_path_lengths, cumulative_distribution, marker="o", linestyle="-")
plt.title("Cumulative Distribution of Average Shortest Path Lengths")
plt.xlabel("Average Shortest Path Length")
plt.ylabel("Cumulative Distribution")
plt.grid(True)
plt.show()

print(dict_sort_by_value(dict(G.degree)))
draw(G, pos, dict(G.degree()), 'Node degree')


degrees = [deg for node, deg in G.degree()]


degree_counts = {}
for degree in degrees:
    if degree in degree_counts:
        degree_counts[degree] += 1
    else:
        degree_counts[degree] = 1


degree_values = list(degree_counts.keys())
node_counts = list(degree_counts.values())


plt.bar(degree_values, node_counts, align='center')
plt.xlabel('Degree')
plt.ylabel('Number of Nodes')
plt.title('Node Degree Distribution')
plt.show()


# print(nx.degree_centrality(G))
# draw(G, pos, nx.degree_centrality(G), 'degree centrality')


# print(nx.eigenvector_centrality(G, max_iter=1000))
# draw(G, pos, nx.eigenvector_centrality(G, max_iter=1000), 'eigenvector centrality')


# print(nx.betweenness_centrality(G))
# draw(G, pos, nx.betweenness_centrality(G), 'betweenness centrality')


# print(nx.closeness_centrality(G))
# draw(G, pos, nx.closeness_centrality(G), 'clossness centrality')


# print(dict_sort_by_value(nx.pagerank(G)))
# draw(G, pos, nx.pagerank(G, alpha=0.85), 'PageRank')

# Katz Centrality
# print(dict_sort_by_value(nx.katz_centrality(G, alpha=0.1, beta=1.0)))
# draw(G, pos, nx.katz_centrality(G, alpha=0.1, beta=1.0), 'Katz Centrality')

# HITS Hubs and Authorities
# h, a=nx.hits(G)
# draw(G, pos, h, 'DiGraph HITS Hubs')
# draw(G, pos, a, 'DiGraph HITS Authorities')


# print(dict_sort_by_value(nx.triangles(G)))
# draw(G, pos, nx.triangles(G), 'Triangles')


# print(dict_sort_by_value(nx.clustering(G)))
# draw(G, pos, nx.clustering(G), 'Clustering Coeffcient')
def evaluate_robustness(G):
    num_nodes = len(G.nodes)
    robustness = []

    for i in range(1, num_nodes + 1):
        # 删除i个随机节点
        nodes_to_remove = np.random.choice(G.nodes, size=i, replace=False)
        G_copy = G.copy()
        G_copy.remove_nodes_from(nodes_to_remove)
        # 计算连通分量数
        num_components = nx.number_connected_components(G_copy)
        robustness.append(num_components)

    return robustness



def evaluate_recoverability(G, removed_nodes):
    recoverability = []

    for node in removed_nodes:
        G_copy = G.copy()
        G_copy.add_node(node)
        G_copy.add_edges_from((node, n) for n in G.neighbors(node) if G.has_edge(node, n))
        num_components = nx.number_connected_components(G_copy)
        recoverability.append(num_components)

    return recoverability



def evaluate_vulnerability(G):
    vulnerability = []
    nodes = list(G.nodes)

    for node in nodes:
        G_copy = G.copy()
        G_copy.remove_node(node)
        num_components = nx.number_connected_components(G_copy)
        vulnerability.append(num_components)

    return vulnerability



def plot_network_metrics(G):
    nodes = list(G.nodes)


    robustness = evaluate_robustness(G)
    plt.figure(figsize=(15, 5))
    plt.subplot(1, 3, 1)
    plt.plot(range(1, len(robustness) + 1), robustness, marker='o', color='r')
    plt.title('Robustness')
    plt.xlabel('Nodes Removed')
    plt.ylabel('Number of Components')


    removed_nodes = np.random.choice(G.nodes, size=3, replace=False)
    recoverability = evaluate_recoverability(G, removed_nodes)
    plt.subplot(1, 3, 2)
    plt.plot(removed_nodes, recoverability, marker='o', color='g')
    plt.title('Recoverability')
    plt.xlabel('Nodes Removed')
    plt.ylabel('Number of Components')


    vulnerability = evaluate_vulnerability(G)
    plt.subplot(1, 3, 3)
    plt.plot(nodes, vulnerability, marker='o', color='b')
    plt.title('Vulnerability')
    plt.xlabel('Nodes')
    plt.ylabel('Number of Components')

    plt.tight_layout()
    plt.show()



plot_network_metrics(G)